/* Name: main.c
 * Project: TravelBug
 * Author: Flip van den Berg - www.flipwork.nl
 * Creation Date: 2008-10-06
 * Based on AVR-USB drivers from Objective Developments - http://www.obdev.at/products/avrusb/index.html
 * Extended: Jindra Fučík - www.fucik.name
 */

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdlib.h>

#include "usbdrv.h"
#include "oddebug.h"

#define BUTTON_PORT PORTB       /* PORTx - register for button output */
#define BUTTON_PIN PINB         /* PINx - register for button input */
#define BUTTON_BIT PB3          /* bit for button input/output */

/* ------------------------------------------------------------------------- */

static uchar    reportBuffer[2];    /* buffer for HID reports */
static uchar    idleRate;           /* in 4 ms units */
static uchar    reportCount;		/* current report */

static uchar	debounceTimeIsOver;	/* for switch debouncing */

static int		MsgCnt;		/*  character count in msg */

/* ------------------------------------------------------------------------- */

PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = { /* USB report descriptor */
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};
/* We use a simplifed keyboard report descriptor which does not support the
 * boot protocol. We don't allow setting status LEDs and we only allow one
 * simultaneous key press (except modifiers). We can therefore use short
 * 2 byte input reports.
 * The report descriptor has been created with usb.org's "HID Descriptor Tool"
 * which can be downloaded from http://www.usb.org/developers/hidpage/.
 * Redundant entries (such as LOGICAL_MINIMUM and USAGE_PAGE) have been omitted
 * for the second INPUT item.
 */
 
PROGMEM char asciitable[256] = {
	0x00, 0x00,	//0 NUL
	0x00, 0x00,	//1 SOH (start of heading)
	0x00, 0x00,	//2 STX (start of text)
	0x00, 0x00,	//3 ETX (end of text)
	0x00, 0x00,	//4 EOT (end of transmission)
	0x00, 0x00,	//5 ENQ (enquiry)
	0x00, 0x00,	//6 ACK (acknowledge)
	0x00, 0x00,	//7 BEL '\a' (bell)
	0x00, 0x2A,	//8 BS  '\b' (backspace)
	0x00, 0x2B,	//9 HT  '\t' (horizontal tab)
	0x00, 0x00,	//10 LF  '\n' (new line)
	0x00, 0x00,	//11 VT  '\v' (vertical tab)
	0x00, 0x00,	//12 FF  '\f' (form feed)
	0x00, 0x28,	//13 CR  '\r' (carriage ret)
	0x00, 0x00,	//14 SO  (shift out)
	0x00, 0x00,	//15 SI  (shift in)
	0x00, 0x00,	//16 DLE (data link escape)
	0x00, 0x00,	//17 DC1 (device control 1)
	0x00, 0x00,	//18 DC2 (device control 2)
	0x00, 0x00,	//19 DC3 (device control 3)
	0x00, 0x00,	//20 DC4 (device control 4)
	0x00, 0x00,	//21 NAK (negative ack.)
	0x00, 0x00,	//22 SYN (synchronous idle)
	0x00, 0x00,	//23 ETB (end of trans. blk)
	0x00, 0x00,	//24 CAN (cancel)
	0x00, 0x00,	//25 EM  (end of medium)
	0x00, 0x00,	//26 SUB (substitute)
	0x00, 0x29,	//27 ESC (escape)
	0x00, 0x00,	//28 FS  (file separator)
	0x00, 0x00,	//29 GS  (group separator)
	0x00, 0x00,	//30 RS  (record separator)
	0x00, 0x00,	//31 US  (unit separator)
	0x00, 0x2C,	//32 SPACE
	0x02, 0x1E,	//33 !
	0x02, 0x34,	//34 "
	0x02, 0x20,	//35 #
	0x02, 0x21,	//36 $
	0x02, 0x22,	//37 %
	0x02, 0x24,	//38 &
	0x00, 0x34,	//39 '
	0x02, 0x26,	//40 (
	0x02, 0x27,	//41 )
	0x02, 0x25,	//42 *
	0x02, 0x2E,	//43 +
	0x00, 0x36,	//44 ,
	0x00, 0x2D,	//45 -
	0x00, 0x37,	//46 .
	0x00, 0x38,	//47 /
	0x00, 0x27,	//48 0
	0x00, 0x1E,	//49 1
	0x00, 0x1F,	//50 2
	0x00, 0x20,	//51 3
	0x00, 0x21,	//52 4
	0x00, 0x22,	//53 5
	0x00, 0x23,	//54 6
	0x00, 0x24,	//55 7
	0x00, 0x25,	//56 8
	0x00, 0x26,	//57 9
	0x02, 0x33,	//58 :
	0x00, 0x33,	//59 ;
	0x02, 0x36,	//60 <
	0x00, 0x2E,	//61 =
	0x02, 0x37,	//62 >
	0x02, 0x38,	//63 ?
	0x02, 0x1F,	//64 @
	0x02, 0x04,	//65 A
	0x02, 0x05,	//66 B
	0x02, 0x06,	//67 C
	0x02, 0x07,	//68 D
	0x02, 0x08,	//69 E
	0x02, 0x09,	//70 F
	0x02, 0x0A,	//71 G
	0x02, 0x0B,	//72 H
	0x02, 0x0C,	//73 I
	0x02, 0x0D,	//74 J
	0x02, 0x0E,	//75 K
	0x02, 0x0F,	//76 L
	0x02, 0x10,	//77 M
	0x02, 0x11,	//78 N
	0x02, 0x12,	//79 O
	0x02, 0x13,	//80 P
	0x02, 0x14,	//81 Q
	0x02, 0x15,	//82 R
	0x02, 0x16,	//83 S
	0x02, 0x17,	//84 T
	0x02, 0x18,	//85 U
	0x02, 0x19,	//86 V
	0x02, 0x1A,	//87 W
	0x02, 0x1B,	//88 X
	0x02, 0x1C,	//89 Y
	0x02, 0x1D,	//90 Z
	0x00, 0x2F,	//91 hranata otevrit
	0x00, 0x31,	//92 obracene lomitko
	0x00, 0x30,	//93 hranata zavrit
	0x02, 0x23,	//94 ^
	0x02, 0x2D,	//95 _
	0x00, 0x35,	//96 `
	0x00, 0x04,	//97 a
	0x00, 0x05,	//98 b
	0x00, 0x06,	//99 c
	0x00, 0x07,	//100 d
	0x00, 0x08,	//101 e
	0x00, 0x09,	//102 f
	0x00, 0x0A,	//103 g
	0x00, 0x0B,	//104 h
	0x00, 0x0C,	//105 i
	0x00, 0x0D,	//106 j
	0x00, 0x0E,	//107 k
	0x00, 0x0F,	//108 l
	0x00, 0x10,	//109 m
	0x00, 0x11,	//110 n
	0x00, 0x12,	//111 o
	0x00, 0x13,	//112 p
	0x00, 0x14,	//113 q
	0x00, 0x15,	//114 r
	0x00, 0x16,	//115 s
	0x00, 0x17,	//116 t
	0x00, 0x18,	//117 u
	0x00, 0x19,	//118 v
	0x00, 0x1A,	//119 w
	0x00, 0x1B,	//120 x
	0x00, 0x1C,	//121 y
	0x00, 0x1D,	//122 z
	0x02, 0x2F,	//123 {
	0x02, 0x31,	//124 |
	0x02, 0x30,	//125 }
	0x02, 0x35,	//126 ~
	0x00, 0x4C	//127 DEL
};
	
#define xTerm 0xFF
#define xNull 0x00


uchar EEMEM ee_OSCCAL = 0xff;
uchar EEMEM msg[] = { '\r','\r','\r','\r',
'D','o','b','r','y',' ','d','e','n',',','\r',
'd','o','s','t','a','l',' ','s','e',' ','v','a','m',' ','d','o',' ','r','u','k','o','u',' ','t','r','a','v','e','l',' ','b','u','g',' ','A','m','a','r','o','.','\r',
'P','r','e','d','p','l','a','t','n','e',' ','n','a',':',' ','h','t','t','p',':','/','/','w','w','w','.','a','m','a','r','o','.','c','z','/','\r',
'\r',
'D','e','k','u','j','i','\r',
'J','i','n','d','r','a',' ','h','t','t','p',':','/','/','w','w','w','.','f','u','c','i','k','.','n','a','m','e','/','T','B','/','\r',
xTerm
};
/*
Dobry den,
dostal se vam do rukou travel bug Amaro.
Predplatne na: http://www.amaro.cz/
Dekuji
Jindra http://www.fucik.name/
*/
static void timerPoll(void)
{
	static unsigned int timerCnt;

    if(TIFR & (1 << TOV1)){
        TIFR = (1 << TOV1); /* clear overflow */
        if (((MsgCnt>3) && (++timerCnt >= 20)) || (++timerCnt >= 126)){       // 22/63 sec delay for next key, 126/63 sec for first key to stabilize
			timerCnt = 0;
			debounceTimeIsOver = 1; 
        }
    }
}

static void buildReport(void)
{
uchar mychar = 0xff;
uchar key = 0; //if not changed by the if-statement below, then send an empty report
uchar mdfr = 0; //if not changed by the if-statement below, then send an empty report

    if(reportCount == 0){
		mychar = eeprom_read_byte(&msg[MsgCnt++]); // read character
		if (mychar > 127) {	// terminator found
			MsgCnt=-1;
			mdfr=0;
			key=0;
		} else {
			mdfr=pgm_read_byte_near(&(asciitable[mychar*2]));
			key=pgm_read_byte_near(&(asciitable[mychar*2 + 1]));
		}

    }

	reportCount++;
    reportBuffer[0] = mdfr;    /* no modifiers */
    reportBuffer[1] = key;
}

static void checkButtonChange(void) {
	
	//uchar tempButtonValue = bit_is_clear(BUTTON_PIN, BUTTON_BIT); //status of switch is stored in tempButtonValue 

	if ((debounceTimeIsOver == 1) && (MsgCnt >= 0)){ //if status has changed and the debounce-delay is over
		//buttonState = tempButtonValue;	// change buttonState to new state
		debounceTimeIsOver = 0;	// debounce timer starts
		reportCount = 0; // start report 
	}
}

/* ------------------------------------------------------------------------- */

static void timerInit(void)
{
    TCCR1 = 0x0b;           /* select clock: 16.5M/1k -> overflow rate = 16.5M/256k = 62.94 Hz */
}

/* -------------------------------------------------------------------------------- */
/* ------------------------ interface to USB driver ------------------------ */
/* -------------------------------------------------------------------------------- */

uchar	usbFunctionSetup(uchar data[8])
{
usbRequest_t    *rq = (void *)data;

    usbMsgPtr = reportBuffer;
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
            /* we only have one report type, so don't look at wValue */
            buildReport();
            return sizeof(reportBuffer);
        }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
            usbMsgPtr = &idleRate;
            return 1;
        }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
            idleRate = rq->wValue.bytes[1];
        }
    }else{
        /* no vendor specific requests implemented */
    }
	return 0;
}

/* ------------------------------------------------------------------------- */
/* ------------------------ Oscillator Calibration ------------------------- */
/* ------------------------------------------------------------------------- */

/* Calibrate the RC oscillator to 8.25 MHz. The core clock of 16.5 MHz is
 * derived from the 66 MHz peripheral clock by dividing. Our timing reference
 * is the Start Of Frame signal (a single SE0 bit) available immediately after
 * a USB RESET. We first do a binary search for the OSCCAL value and then
 * optimize this value with a neighboorhod search.
 * This algorithm may also be used to calibrate the RC oscillator directly to
 * 12 MHz (no PLL involved, can therefore be used on almost ALL AVRs), but this
 * is wide outside the spec for the OSCCAL value and the required precision for
 * the 12 MHz clock! Use the RC oscillator calibrated to 12 MHz for
 * experimental purposes only!
 */
static void calibrateOscillator(void)
{
uchar       step = 128;
uchar       trialValue = 0, optimumValue;
int         x, optimumDev, targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);

    /* do a binary search: */
    do{
        OSCCAL = trialValue + step;
        x = usbMeasureFrameLength();    /* proportional to current real frequency */
        if(x < targetValue)             /* frequency still too low */
            trialValue += step;
        step >>= 1;
    }while(step > 0);
    /* We have a precision of +/- 1 for optimum OSCCAL here */
    /* now do a neighborhood search for optimum value */
    optimumValue = trialValue;
    optimumDev = x; /* this is certainly far away from optimum */
    for(OSCCAL = trialValue - 1; OSCCAL <= trialValue + 1; OSCCAL++){
        x = usbMeasureFrameLength() - targetValue;
        if(x < 0)
            x = -x;
        if(x < optimumDev){
            optimumDev = x;
            optimumValue = OSCCAL;
        }
    }
    OSCCAL = optimumValue;
}
/*
Note: This calibration algorithm may try OSCCAL values of up to 192 even if
the optimum value is far below 192. It may therefore exceed the allowed clock
frequency of the CPU in low voltage designs!
You may replace this search algorithm with any other algorithm you like if
you have additional constraints such as a maximum CPU clock.
For version 5.x RC oscillators (those with a split range of 2x128 steps, e.g.
ATTiny25, ATTiny45, ATTiny85), it may be useful to search for the optimum in
both regions.
*/

void    usbEventResetReady(void)
{
    calibrateOscillator();
    eeprom_write_byte(&ee_OSCCAL, OSCCAL);   /* store the calibrated value in EEPROM */
}

/* ------------------------------------------------------------------------- */
/* --------------------------------- main ---------------------------------- */
/* ------------------------------------------------------------------------- */

int main(void)
{
uchar   i;
uchar   calibrationValue;

    calibrationValue = eeprom_read_byte(&ee_OSCCAL); /* calibration value from last time */
    if(calibrationValue != 0xff){
        OSCCAL = calibrationValue;
    }
    
	//odDebugInit();
    usbInit();
    usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
    i = 0;
    while(--i){             /* fake USB disconnect for > 250 ms */
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();

    wdt_enable(WDTO_1S);

	/* turn on internal pull-up resistor for the switch */
    BUTTON_PORT |= _BV(BUTTON_BIT);

    timerInit();

    sei();

	MsgCnt =0;
	
    for(;;){    /* main event loop */
        wdt_reset();
        usbPoll();

		/* A USB keypress cycle is defined as a scancode being present in a report, and
		then absent from a later report. To press and release the Caps Lock key, instead of
		holding it down, we need to send the report with the Caps Lock scancode and
		then an empty report. */
		
		if(usbInterruptIsReady() && reportCount < 2){ /* we can send another key */
        	buildReport();
           	usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
        	}
        checkButtonChange();
		timerPoll();
	}
   	return 0;
}
